import csv
import dataclasses
import json
import logging
from datetime import datetime, timezone
from decimal import Decimal

import requests

from pricehist import __version__
from pricehist.price import Price

from .basesource import BaseSource


class Yahoo(BaseSource):
    def id(self):
        return "yahoo"

    def name(self):
        return "Yahoo! Finance"

    def description(self):
        return (
            "Historical data for most Yahoo! Finance symbols, "
            "as available on the web page"
        )

    def source_url(self):
        return "https://finance.yahoo.com/"

    def start(self):
        return "1970-01-01"

    def types(self):
        return ["adjclose", "open", "high", "low", "close", "mid"]

    def notes(self):
        return (
            "Yahoo! Finance decommissioned its historical data API in 2017 but "
            "some historical data is available via its web page, as described in: "
            "https://help.yahoo.com/kb/"
            "download-historical-data-yahoo-finance-sln2311.html\n"
            f"{self._symbols_message()}\n"
            "In output the base and quote will be the Yahoo! symbol and its "
            "corresponding currency. Some symbols include the name of the quote "
            "currency (e.g. BTC-USD), so you may wish to use --fmt-base to "
            "remove the redundant information.\n"
            "When a symbol's historical data is unavilable due to data licensing "
            "restrictions, its web page will show no download button and "
            "pricehist will only find the current day's price."
        )

    def _symbols_message(self):
        return (
            "Find the symbol of interest on https://finance.yahoo.com/ and use "
            "that as the PAIR in your pricehist command. Prices for each symbol "
            "are given in its native currency."
        )

    def symbols(self):
        logging.info(self._symbols_message())
        return []

    def fetch(self, series):
        # TODO fail if quote isn't empty - yahoo symbols don't have a slash
        spark, history = self._data(series)

        output_quote = spark["spark"]["result"][0]["response"][0]["meta"]["currency"]

        prices = [
            Price(row["date"], amount)
            for row in history
            if (amount := self._amount(row, series.type))
        ]

        return dataclasses.replace(series, quote=output_quote, prices=prices)

    def _amount(self, row, type):
        if type != "mid" and row[type] != "null":
            return Decimal(row[type])
        elif type == "mid" and row["high"] != "null" and row["low"] != "null":
            return sum([Decimal(row["high"]), Decimal(row["low"])]) / 2
        else:
            return None

    def _data(self, series) -> (dict, csv.DictReader):
        base_url = "https://query1.finance.yahoo.com/v7/finance"
        headers = {"User-Agent": f"pricehist/{__version__}"}

        spark_url = f"{base_url}/spark"
        spark_params = {
            "symbols": series.base,
            "range": "1d",
            "interval": "1d",
            "indicators": "close",
            "includeTimestamps": "false",
            "includePrePost": "false",
        }
        spark_response = self.log_curl(
            requests.get(spark_url, params=spark_params, headers=headers)
        )
        spark = json.loads(spark_response.content)

        start_ts = int(
            datetime.strptime(series.start, "%Y-%m-%d")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        )
        end_ts = int(
            datetime.strptime(series.end, "%Y-%m-%d")
            .replace(tzinfo=timezone.utc)
            .timestamp()
        ) + (
            24 * 60 * 60
        )  # round up to include the last day

        history_url = f"{base_url}/download/{series.base}"
        history_params = {
            "period1": start_ts,
            "period2": end_ts,
            "interval": "1d",
            "events": "history",
            "includeAdjustedClose": "true",
        }
        history_response = self.log_curl(
            requests.get(history_url, params=history_params, headers=headers)
        )
        history_lines = history_response.content.decode("utf-8").splitlines()
        history_lines[0] = history_lines[0].lower().replace(" ", "")
        history = csv.DictReader(history_lines, delimiter=",")

        return (spark, history)



from .alphavantage import AlphaVantage
from .coindesk import CoinDesk
from .coinmarketcap import CoinMarketCap
from .ecb import ECB
from .yahoo import Yahoo

by_id = {
    source.id(): source
    for source in [AlphaVantage(), CoinDesk(), CoinMarketCap(), ECB(), Yahoo()]
}


def formatted():
    width = max([len(k) for k, v in by_id.items()])
    lines = [k.ljust(width + 4) + v.name() for k, v in by_id.items()]
    return "\n".join(lines)



import csv
import dataclasses
import json
import logging
import os
from datetime import datetime, timedelta
from decimal import Decimal

import requests

from pricehist.price import Price

from .basesource import BaseSource


class AlphaVantage(BaseSource):
    QUERY_URL = "https://www.alphavantage.co/query"

    def id(self):
        return "alphavantage"

    def name(self):
        return "Alpha Vantage"

    def description(self):
        return "Provider of market data for stocks, forex and cryptocurrencies"

    def source_url(self):
        return "https://www.alphavantage.co/"

    def start(self):
        return "1995-01-01"

    def types(self):
        return ["close", "open", "high", "low", "adjclose", "mid"]

    def notes(self):
        keystatus = "already set" if self._apikey(require=False) else "NOT YET set"
        return (
            "Alpha Vantage has data on digital (crypto) currencies, physical "
            "(fiat) currencies and stocks.\n"
            "An API key is required. One can be obtained for free from "
            "https://www.alphavantage.co/support/#api-key and should be made "
            "available in the ALPHAVANTAGE_API_KEY environment variable "
            f"({keystatus}).\n"
            "The PAIR for currencies should be in BASE/QUOTE form. The quote "
            "symbol must always be for a physical currency. The --symbols option "
            "will list all digital and physical currency symbols.\n"
            "The PAIR for stocks is the stock symbol only. The quote currency "
            f"will be determined automatically. {self._stock_symbols_message()}\n"
            "The price type 'adjclose' is only available for stocks.\n"
            "Alpha Vantage's standard API call frequency limits is 5 calls per "
            "minute and 500 per day, so you may need to pause between successive "
            "commands. Note that retrieving prices for one stock requires two "
            "calls."
        )

    def _stock_symbols_message(self):
        return "Stock symbols can be discovered using the --search option."

    def symbols(self):
        logging.info(self._stock_symbols_message())
        return self._digital_symbols() + self._physical_symbols()

    def search(self, query):
        data = self._search_data(query)
        results = [
            (
                m["1. symbol"],
                ", ".join(
                    [
                        m["2. name"],
                        m["3. type"],
                        m["4. region"],
                        m["8. currency"],
                    ]
                ),
            )
            for m in data["bestMatches"]
        ]
        return results

    def fetch(self, series):
        output_base = series.base.upper()
        output_quote = series.quote

        if series.quote == "":
            output_quote = self._stock_currency(output_base)
            data = self._stock_data(series)
        else:
            if series.type == "adjclose":
                logging.critical(
                    "The 'adjclose' price type is only available for stocks. "
                    "Use 'close' instead."
                )
                exit(1)
            elif series.base in [s for s, n in self._physical_symbols()]:
                data = self._physical_data(series)
            else:
                data = self._digital_data(series)

        prices = [
            Price(day, amount)
            for day, entries in data.items()
            if (amount := self._amount(day, entries, series))
        ]

        return dataclasses.replace(
            series, base=output_base, quote=output_quote, prices=prices
        )

    def _amount(self, day, entries, series):
        if day < series.start or day > series.end:
            return None
        elif type == "mid":
            return sum([Decimal(entries["high"]), Decimal(entries["low"])]) / 2
        else:
            return Decimal(entries[series.type])

    def _stock_currency(self, symbol):
        data = self._search_data(symbol)
        for match in data["bestMatches"]:
            if match["1. symbol"] == symbol:
                return match["8. currency"]
        return "Unknown"

    def _search_data(self, keywords: str):
        params = {
            "function": "SYMBOL_SEARCH",
            "keywords": keywords,
            "apikey": self._apikey(),
        }
        response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        data = json.loads(response.content)
        return data

    def _stock_data(self, series):
        params = {
            "function": "TIME_SERIES_DAILY_ADJUSTED",
            "symbol": series.base,
            "outputsize": self._outputsize(series.start),
            "apikey": self._apikey(),
        }
        response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        data = json.loads(response.content)
        normalized_data = {
            day: {
                "open": entries["1. open"],
                "high": entries["2. high"],
                "low": entries["3. low"],
                "close": entries["4. close"],
                "adjclose": entries["5. adjusted close"],
            }
            for day, entries in reversed(data["Time Series (Daily)"].items())
        }
        return normalized_data

    def _physical_data(self, series):
        params = {
            "function": "FX_DAILY",
            "from_symbol": series.base,
            "to_symbol": series.quote,
            "outputsize": self._outputsize(series.start),
            "apikey": self._apikey(),
        }
        response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        data = json.loads(response.content)
        normalized_data = {
            day: {k[3:]: v for k, v in entries.items()}
            for day, entries in reversed(data["Time Series FX (Daily)"].items())
        }
        return normalized_data

    def _outputsize(self, start):
        almost_100_days_ago = (datetime.now().date() - timedelta(days=95)).isoformat()
        if start < almost_100_days_ago:
            return "full"
        else:
            return "compact"

    def _digital_data(self, series):
        params = {
            "function": "DIGITAL_CURRENCY_DAILY",
            "symbol": series.base,
            "market": series.quote,
            "apikey": self._apikey(),
        }
        response = self.log_curl(requests.get(self.QUERY_URL, params=params))
        data = json.loads(response.content)
        normalized_data = {
            day: {
                "open": entries[f"1a. open ({series.quote})"],
                "high": entries[f"2a. high ({series.quote})"],
                "low": entries[f"3a. low ({series.quote})"],
                "close": entries[f"4a. close ({series.quote})"],
            }
            for day, entries in reversed(
                # TODO key error here when body is
                # { "Error Message": "Invalid API call. ...." }
                # can be due to asking for stock data with a quote currency
                # error message: no digital currency, if it's stock remove quote
                data["Time Series (Digital Currency Daily)"].items()
            )
        }
        return normalized_data

    def _apikey(self, require=True):
        key_name = "ALPHAVANTAGE_API_KEY"
        key = os.getenv(key_name)
        if require and not key:
            logging.critical(
                f"The environment variable {key_name} is empty. "
                "Get a free API key from https://www.alphavantage.co/support/#api-key, "
                f'export {key_name}="YOUR_OWN_API_KEY" and retry.'
            )
            exit(1)
        return key

    def _physical_symbols(self) -> list[(str, str)]:
        url = "https://www.alphavantage.co/physical_currency_list/"
        response = self.log_curl(requests.get(url))
        lines = response.content.decode("utf-8").splitlines()
        data = csv.reader(lines[1:], delimiter=",")
        return [(s, f"Physical: {n}") for s, n in data]

    def _digital_symbols(self) -> list[(str, str)]:
        url = "https://www.alphavantage.co/digital_currency_list/"
        response = self.log_curl(requests.get(url))
        lines = response.content.decode("utf-8").splitlines()
        data = csv.reader(lines[1:], delimiter=",")
        return [(s, f"Digital: {n}") for s, n in data]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pricehist', 'pricehist.outputs', 'pricehist.resources', 'pricehist.sources']

package_data = \
{'': ['*']}

install_requires = \
['cssselect>=1.1.0,<2.0.0',
 'curlify>=2.2.1,<3.0.0',
 'lxml>=4.6.2,<5.0.0',
 'requests>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['pricehist = pricehist.cli:cli']}

setup_kwargs = {
    'name': 'pricehist',
    'version': '0.1.5',
    'description': 'Fetch and format historical price data',
    'long_description': "# pricehist\n\nA command-line tool for fetching and formatting historical price data, with\nsupport for multiple data sources and output formats.\n\n## Installation\n\nInstall via [pip](https://pip.pypa.io/en/stable/) or\n[pipx](https://pypa.github.io/pipx/):\n\n```bash\npipx install pricehist\n```\n\n## Sources\n\n- **`alphavantage`**: [Alpha Vantage](https://www.alphavantage.co/)\n- **`coindesk`**: [CoinDesk Bitcoin Price Index](https://www.coindesk.com/coindesk-api)\n- **`coinmarketcap`**: [CoinMarketCap](https://coinmarketcap.com/)\n- **`ecb`**: [European Central Bank Euro foreign exchange reference rates](https://www.ecb.europa.eu/stats/exchange/eurofxref/html/index.en.html)\n- **`yahoo`**: [Yahoo! Finance](https://finance.yahoo.com/)\n\n## Output formats\n\n- **`beancount`**: [Beancount](http://furius.ca/beancount/)\n- **`csv`**: [Comma-separated values](https://en.wikipedia.org/wiki/Comma-separated_values)\n- **`gnucash-sql`**: [GnuCash](https://www.gnucash.org/) SQL\n- **`ledger`**: [Ledger](https://www.ledger-cli.org/) and [hledger](https://hledger.org/)\n\n## Examples\n\nShow usage information:\n\n```bash\npricehist -h\n```\n```\nusage: pricehist [-h] [--version] [--verbose] {sources,source,fetch} ...\n\nFetch historical price data\n\noptional arguments:\n  -h, --help              show this help message and exit\n  --version               show version information\n  --verbose               show all log messages\n\ncommands:\n  {sources,source,fetch}\n    sources               list sources\n    source                show source details\n    fetch                 fetch prices\n```\n\nShow usage information for the `fetch` command:\n\n```\npricehist fetch -h\n```\n```\nusage: pricehist fetch SOURCE PAIR [-h] [-t TYPE] [-s DATE | -sx DATE] [-e DATE | -ex DATE]\n[-o beancount|csv|gnucash-sql|ledger] [--invert] [--quantize INT]\n[--fmt-base SYM] [--fmt-quote SYM] [--fmt-time TIME] [--fmt-decimal CHAR] [--fmt-thousands CHAR]\n[--fmt-symbol rightspace|right|leftspace|left] [--fmt-datesep CHAR] [--fmt-csvdelim CHAR]\n\npositional arguments:\n  SOURCE                   the source identifier\n  PAIR                     pair, usually BASE/QUOTE, e.g. BTC/USD\n\noptional arguments:\n  -h, --help               show this help message and exit\n  -t TYPE, --type TYPE     price type, e.g. close\n  -s DATE, --start DATE    start date, inclusive (default: source start)\n  -sx DATE, --startx DATE  start date, exclusive\n  -e DATE, --end DATE      end date, inclusive (default: today)\n  -ex DATE, --endx DATE    end date, exclusive\n  -o FMT, --output FMT     output format (default: csv)\n  --invert                 invert the price, swapping base and quote\n  --quantize INT           round to the given number of decimal places\n  --fmt-base SYM           rename the base symbol in output\n  --fmt-quote SYM          rename the quote symbol in output\n  --fmt-time TIME          set a particular time of day in output (default: 00:00:00)\n  --fmt-decimal CHAR       decimal point in output (default: '.')\n  --fmt-thousands CHAR     thousands separator in output (default: '')\n  --fmt-symbol LOCATION    commodity symbol placement in output (default: rightspace)\n  --fmt-datesep CHAR       date separator in output (default: '-')\n  --fmt-csvdelim CHAR      field delimiter for CSV output (default: ',')\n```\n\nFetch prices after 2021-01-04, ending 2021-01-15, as CSV:\n\n```bash\npricehist fetch ecb EUR/AUD -sx 2021-01-04 -e 2021-01-15 -o csv\n```\n```\ndate,base,quote,amount,source,type\n2021-01-05,EUR,AUD,1.5927,ecb,reference\n2021-01-06,EUR,AUD,1.5824,ecb,reference\n2021-01-07,EUR,AUD,1.5836,ecb,reference\n2021-01-08,EUR,AUD,1.5758,ecb,reference\n2021-01-11,EUR,AUD,1.5783,ecb,reference\n2021-01-12,EUR,AUD,1.5742,ecb,reference\n2021-01-13,EUR,AUD,1.5734,ecb,reference\n2021-01-14,EUR,AUD,1.5642,ecb,reference\n2021-01-15,EUR,AUD,1.568,ecb,reference\n```\n\nIn Ledger format:\n\n```bash\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o ledger | head\n```\n```\nP 2021-01-04 00:00:00 EUR 1.5928 AUD\nP 2021-01-05 00:00:00 EUR 1.5927 AUD\nP 2021-01-06 00:00:00 EUR 1.5824 AUD\nP 2021-01-07 00:00:00 EUR 1.5836 AUD\nP 2021-01-08 00:00:00 EUR 1.5758 AUD\nP 2021-01-11 00:00:00 EUR 1.5783 AUD\nP 2021-01-12 00:00:00 EUR 1.5742 AUD\nP 2021-01-13 00:00:00 EUR 1.5734 AUD\nP 2021-01-14 00:00:00 EUR 1.5642 AUD\nP 2021-01-15 00:00:00 EUR 1.568 AUD\n```\n\nGenerate SQL for a GnuCash database and apply it immediately:\n\n```bash\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | sqlite3 Accounts.gnucash\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | mysql -u username -p -D databasename\npricehist fetch ecb EUR/AUD -s 2021-01-01 -o gnucash-sql | psql -U username -d databasename -v ON_ERROR_STOP=1\n```\n\n## Design choices\n\nTo keep things simple, at least for now, `pricehist` provides only univariate\ntime series of daily historical prices. It doesn't provide other types of\nmarket, financial or economic data, real-time prices, or other temporal\nresolutions. Multiple or multivariate series require multiple invocations.\n\n## Alternatives\n\nBeancount's [`bean-price`](https://github.com/beancount/beanprice) tool fetches\nprices and addresses other workflow concerns in a Beancount-specific manner,\ngenerally requiring a Beancount file as input.\n\nThe [Piecash](https://piecash.readthedocs.io/) library is a pythonic interface\nto GnuCash files stored in SQL which has a\n[`Commodity.update_prices`](https://piecash.readthedocs.io/en/master/api/piecash.core.commodity.html?highlight=update_prices#piecash.core.commodity.Commodity.update_prices)\nmethod for fetching historical prices.\nThe GnuCash wiki documents [wrapper scripts](https://wiki.gnucash.org/wiki/Stocks/get_prices)\nfor the [Finance::QuoteHist](https://metacpan.org/pod/Finance::QuoteHist) Perl\nmodule.\n\nOther projects with related goals include:\n\n* [`hledger-stockquotes`](https://github.com/prikhi/hledger-stockquotes):\n  A CLI addon for hledger that reads a journal file and pulls the historical prices for commodities.\n* [`ledger_get_prices`](https://github.com/nathankot/ledger-get-prices):\n  Uses Yahoo Finance to generate a price database based on your current Ledger commodities and time period.\n* [LedgerStockUpdate](https://github.com/adchari/LedgerStockUpdate):\n  Locates any stocks you have in your ledger-cli file, then generates a price database of those stocks.\n* [`market-prices`](https://github.com/barrucadu/hledger-scripts#market-prices):\n  Downloads market values of commodities from a few different sources.\n",
    'author': 'Chris Berkhout',
    'author_email': 'chris@chrisberkhout.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/chrisberkhout/pricehist',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
